__all__ = ['Process', 'Interaction', 'GeneticInteraction',
           'MolecularInteraction', 'TemplateReaction',
           'Control', 'Conversion', 'Catalysis',
           'TemplateReactionRegulation', 'Modulation',
           'ComplexAssembly', 'BiochemicalReaction',
           'Degradation', 'Transport', 'TransportWithBiochemicalReaction']

from .base import Entity


class Process(Entity):
    """BioPAX Process."""
    def __init__(self,
                 **kwargs):
        super().__init__(**kwargs)
        self._controlled_of = set()
        self._step_process_of = set()
        self._pathway_component_of = set()

    @property
    def controlled_of(self):
        return self._controlled_of

    @property
    def step_process_of(self):
        return self._step_process_of

    @property
    def pathway_component_of(self):
        return self._pathway_component_of


class Interaction(Process):
    """BioPAX Interaction.

    Attributes
    ----------
    participant : List[Entity]
    interaction_type : List[str]
    """
    list_types = Process.list_types + ['participant', 'interaction_type']

    def __init__(self,
                 participant=None,
                 interaction_type=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.participant = participant if participant else []
        self.interaction_type = interaction_type if interaction_type else []


class GeneticInteraction(Interaction):
    """BioPAX GeneticInteraction."""
    pass


class MolecularInteraction(Interaction):
    """BioPAX MolecularInteraction."""
    pass


class TemplateReaction(Interaction):
    """BioPAX TemplateReaction.

    Attributes
    ----------
    template : NucleicAcid
    product : List[PhysicalEntity]
    template_direction : str
    """
    list_types = Interaction.list_types + ['product']

    def __init__(self,
                 template=None,
                 product=None,
                 template_direction=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.template = template
        self.product = product if product else []
        self.template_direction = template_direction


class Control(Interaction):
    """BioPAX Control.

    Attributes
    ----------
    control_type : str
    controller : List[Process]
    controlled : Process
    """
    list_types = Interaction.list_types + ['controller']

    def __init__(self,
                 control_type=None,
                 controller=None,
                 controlled=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.control_type = control_type
        self.controller = controller if controller else []
        self.controlled = controlled


class Conversion(Interaction):
    """BioPAX Conversion.

    Attributes
    ----------
    left : List[PhysicalEntity]
    right : List[PhysicalEntity]
    conversion_direction : str
    participant_stoichiometry : List[Stoichiometry]
    spontaneous : bool
    """
    list_types = Interaction.list_types + \
        ['left', 'right', 'participant_stoichiometry']

    def __init__(self,
                 left=None,
                 right=None,
                 conversion_direction=None,
                 participant_stoichiometry=None,
                 spontaneous=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.left = left if left else []
        self.right = right if right else []
        self.conversion_direction = conversion_direction
        self.participant_stoichiometry = participant_stoichiometry  if \
            participant_stoichiometry else []
        self.spontaneous = spontaneous


class Catalysis(Control):
    """BioPAX Catalysis.

    Attributes
    ----------
    catalysis_direction : str
    cofactor : List[PhysicalEntity]
    """
    list_types = Control.list_types + ['cofactor']

    def __init__(self,
                 catalysis_direction=None,
                 cofactor=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.catalysis_direction = catalysis_direction
        self.cofactor = cofactor if cofactor else []


class TemplateReactionRegulation(Control):
    """BioPAX TemplateReactionRegulation."""
    pass


class Modulation(Control):
    """BioPAX Modulation."""
    pass


class ComplexAssembly(Conversion):
    """BioPAX ComplexAssembly."""
    pass


class BiochemicalReaction(Conversion):
    """BioPAX BiochemicalReaction.

    Attributes
    ----------
    delta_s : List[float]
    delta_h : List[float]
    delta_g : List[DeltaG]
    k_e_q : List[KPrime]
    e_c_number : List[str]
    """
    list_types = Conversion.list_types + ['delta_s', 'delta_h', 'delta_g',
                                          'k_e_q', 'e_c_number']

    def __init__(self,
                 delta_s=None,
                 delta_h=None,
                 delta_g=None,
                 k_e_q=None,
                 e_c_number=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.delta_s = delta_s if delta_s else []
        self.delta_h = delta_h if delta_h else []
        self.delta_g = delta_g if delta_g else []
        self.k_e_q = k_e_q if k_e_q else []
        self.e_c_number = e_c_number if e_c_number else []


class Degradation(Conversion):
    """BioPAX Degradation."""
    pass


class Transport(Conversion):
    """BioPAX Transport."""
    pass


class TransportWithBiochemicalReaction(BiochemicalReaction):
    """BioPAX TransportWithBiochemicalReaction."""
    pass




__all__ = ['BioPaxObject', 'Controller', 'Entity', 'Pathway', 'Gene',
           'Unresolved', 'Observable', 'Named', 'XReferrable']

from typing import List, Optional, TYPE_CHECKING

from ..xml_util import *

if TYPE_CHECKING:
    from .util import Xref


class Unresolved:
    """A placeholder class used while deserializing BioPAX models."""
    def __init__(self, obj_id):
        self.obj_id = obj_id


class BioPaxObject:
    """Generic BioPAX Object. It is the parent class of all more specific
    BioPAX classes."""
    list_types = ['comment']
    xml_types = {}

    def __init__(self, uid, comment=None, **kwargs):
        # Pass on for cooperative inheritance
        super().__init__(**kwargs)
        self.uid = uid
        self.comment = comment if comment else []

    @classmethod
    def from_xml(cls, element):
        uid = get_id_or_about(element)
        kwargs = {'uid': uid}
        for key in cls.list_types:
            kwargs[key] = []
        for child in element.getchildren():
            key = get_attr_tag(child)
            # In some OWL formats, the child is directly defined
            # under this tag, in that case we directly deserialize it.
            if child.getchildren():
                gchild = child.getchildren()[0]
                obj_cls = globals()[get_tag(gchild)]
                val_to_add = obj_cls.from_xml(gchild)
            # Otherwise, we check if the element is a simple type that we
            # can just get as a text value
            elif (get_datatype(child.attrib) is None
                  and not get_resource(child.attrib)) \
                    or is_datatype(child.attrib, 'xsd', 'string') \
                    or is_datatype(child.attrib, 'xsd', 'int') \
                    or is_datatype(child.attrib, 'xsd', 'float'):
                val_to_add = child.text
            # If neither of the above is the case, then we assume that the
            # element is a reference that is defined in another block
            # somewhere so we treat is as Unresolved until later.
            else:
                res = get_resource(child.attrib)
                val_to_add = Unresolved(res)

            if key in cls.list_types:
                kwargs[key].append(val_to_add)
            else:
                kwargs[key] = val_to_add
        return cls(**kwargs)

    def to_xml(self):
        id_type = 'about' if is_url(self.uid) else 'ID'
        element = makers['bp'](self.__class__.__name__,
                               **{nselem('rdf', id_type): self.uid})
        for attr in [a for a in dir(self)
                     if not a.startswith('_')
                     and a not in {'list_types', 'xml_types',
                                   'to_xml', 'from_xml', 'uid'}]:
            val = getattr(self, attr)
            if val is None:
                continue

            # We have to implement special handling for names to make sure
            # we don't serialize display/standard names here
            if attr == 'name' and isinstance(self, Named):
                val = self.get_plain_names()

            if isinstance(val, list):
                for v in val:
                    child_elem = self._simple_to_xml(attr, v)
                    if child_elem is not None:
                        element.append(child_elem)
            else:
                child_elem = self._simple_to_xml(attr, val)
                if child_elem is not None:
                    element.append(child_elem)
        return element

    def _simple_to_xml(self, attr, val):
        if isinstance(val, BioPaxObject):
            child_elem = makers['bp'](
                snake_to_camel(attr),
                **{nselem('rdf', 'resource'):
                    ('#%s' % val.uid) if not is_url(val.uid) else val.uid}
            )
            return child_elem
        elif isinstance(val, str):
            xml_type = self.xml_types.get(attr, 'string')
            child_elem = makers['bp'](
                snake_to_camel(attr),
                val,
                **{nselem('rdf', 'datatype'): nssuffix('xsd', xml_type)}
            )
            return child_elem
        return None


class XReferrable:
    """A mixin class to add xrefs to a BioPaxObject.

    Attributes
    ----------
    xref : List[Xref]
    """
    list_types = ['xref']

    def __init__(self, xref: Optional[List["Xref"]] = None, **kwargs):
        # Pass on for cooperative inheritance
        super().__init__(**kwargs)
        self.xref = xref if xref else []


class Named(XReferrable):
    """A mixin class to add names to a BioPaxObject.

    Attributes
    ----------
    display_name : str
    standard_name : str
    name : str
    """
    list_types = XReferrable.list_types + ['name']

    def __init__(self, display_name=None, standard_name=None, name=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.display_name = display_name
        self.standard_name = standard_name
        self._name = name if name else []

    @property
    def name(self):
        """All names associated with the object including the standard and
        display name, if available."""
        std_name = [self.standard_name] if self.standard_name else []
        disp_name = [self.display_name] if self.display_name else []
        return std_name + disp_name + self._name

    def get_plain_names(self):
        return self._name


class Observable:
    """A mixin class to add evidence to a BioPaxObject.

    Attributes
    ----------
    evidence : List[Evidence]
    """
    list_types = ['evidence']

    def __init__(self, evidence=None, **kwargs):
        # Pass on for cooperative inheritance
        super().__init__(**kwargs)
        self.evidence = evidence if evidence else []


class Entity(BioPaxObject, Observable, Named):
    """BioPAX Entity.

    Attributes
    ----------
    availability : str
    data_source : List[Provenance]
    """
    list_types = BioPaxObject.list_types + Observable.list_types + \
        Named.list_types + ['data_source']

    def __init__(self,
                 availability=None,
                 data_source=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.availability = availability
        self.data_source = data_source if data_source else []
        self._participant_of = set()

    @property
    def participant_of(self):
        return self._participant_of


class Gene(Entity):
    """BioPAX Gene

    Attributes
    ----------
    organism: BioSource
    """
    def __init__(self, organism, **kwargs):
        super().__init__(**kwargs)
        self.organism = organism


class Controller:
    """BioPAX Controller."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._controller_of = set()

    @property
    def controller_of(self):
        return self._controller_of


class Pathway(Entity, Controller):
    """BioPAX Pathway."""
    list_types = Entity.list_types + ['pathway_component']

    def __init__(self,
                 pathway_component=None,
                 pathway_order=None,
                 organism=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.pathway_component = pathway_component if pathway_component else []
        self.pathway_order = pathway_order
        self.organism = organism


# These are necessary to have the objects in the global
# scope, required for some modes of deserialization
from .interaction import *
from .physical_entity import *
from .util import *



"""This module implements finding paths in a BioPaxModel starting from a
given object using a path constraint string."""
__all__ = ['find_objects', 'BiopaxClassConstraintError']

import logging
from typing import List
from .biopax import *


logger = logging.getLogger(__name__)


def find_objects(start_obj: BioPaxObject, path_str: str) -> List[BioPaxObject]:
    """Return objects matching the given path specification.

    Parameters
    ----------
    start_obj :
        The object to start the search from.

    path_str :
        A path specification string which consists of one or more parts
        separated by /. Each part is the name of an object attribute, and
        can optionally contain a class name as well, separated by : to
        constrain the class of the target of the attribute to consider.
        Optionally, each attribute can also have a * suffix to make the
        search recursive.

    Returns
    -------
    :
        A list of BioPaxObjects satisfying the given path specification.
    """
    # We split off the first part and separate out the rest
    part, rest = path_str.split('/', maxsplit=1) \
        if '/' in path_str else (path_str, None)

    # Handle class constraint
    if ':' in part:
        attribute, class_constraint_str = part.split(':', maxsplit=1)
        try:
            cls = biopax_cls_map[class_constraint_str]
        except KeyError:
            raise BiopaxClassConstraintError(class_constraint_str) from None
    else:
        attribute, cls = part, None

    # Handle recursion marker
    if attribute.endswith('*'):
        attribute = attribute[:-1]
        recursive = True
    else:
        recursive = False

    # Get the attribute we are looking for
    attr_val = getattr(start_obj, attribute, None)
    # We turn the value into a flat list of BioPaxObjects
    val = _get_object_list(attr_val)

    # If this is a recursive part, we run a BFS to get all the downstream
    # objects that can be reached via one or more of the given type of
    # attribute links
    if recursive:
        queue = val[:]
        visited = val[:]
        while queue:
            obj = queue.pop(0)
            obj_val = getattr(obj, attribute, None)
            for child in _get_object_list(obj_val):
                if child not in visited:
                    visited.append(child)
                    queue.append(child)
        val = visited

    # At this point, val is guaranteed to be a list of BioPaxObjects
    results = []
    for v in val:
        if cls and not isinstance(v, cls):
            continue
        if rest:
            results += find_objects(v, rest)
        else:
            results.append(v)
    return results


def _get_object_list(val):
    if isinstance(val, BioPaxObject):
        return [val]
    elif isinstance(val, (list, set)):
        return [v for v in val if isinstance(v, BioPaxObject)]
    else:
        return []


def _make_biopax_cls_map():
    import inspect
    from pybiopax import biopax
    return {k: v for k, v in inspect.getmembers(biopax, inspect.isclass)
            if issubclass(v, BioPaxObject)}


biopax_cls_map = _make_biopax_cls_map()


class BiopaxClassConstraintError(KeyError):
    def __init__(self, cls_str):
        self.cls_str = cls_str

    def __str__(self):
        return f'{self.cls_str} is not a valid BioPAX class name.'



import os
import re
import pybiopax
from pybiopax.biopax import *

here = os.path.dirname(os.path.abspath(__file__))


def test_process_owl():
    test_file = os.path.join(here, 'biopax_test.owl.gz')
    model = pybiopax.model_from_owl_gz(test_file)
    assert len(model.objects) == 58027, len(model.objects)
    assert 'BiochemicalReaction_a75d6aaebf5be718d981d95355ced14a' in \
        model.objects
    bcr = model.objects['BiochemicalReaction_a75d6aaebf5be718d981d95355ced14a']
    assert bcr.display_name == \
        'STATE_TRANSITION_LEFT__Raf-RIGHT__Raf-_r21'
    assert isinstance(bcr.left, list)
    assert isinstance(bcr.left[0], Protein)
    assert isinstance(bcr.right, list)
    assert isinstance(bcr.right[0], Protein)

    pub = model.objects['PublicationXref_http___www_phosphosite_org_'
                        'phosphosite_owl_recid_11958410']
    assert len(pub.comment) == 1
    assert pub.comment[0].startswith('REPLACED')
    assert pub.url[0].startswith('http://www.phosphosite')
    assert pub.year == '2010', pub.year


def test_process_molecular_interactions():
    test_file = os.path.join(here, 'molecular_interactions_test.owl')
    model = pybiopax.model_from_owl_file(test_file)
    assert len(model.objects) == 62
    mol_int = \
        model.objects['MolecularInteraction_1e82d9951c7d71c02ee6e7bdc7cb8e47']
    assert isinstance(mol_int.participant, list)
    assert len(mol_int.participant) == 2
    disp_names = {part.display_name for part in mol_int.participant}
    assert disp_names == {'ALG6', 'ALG8'}
    names = set.union(*[set(part.name) for part in mol_int.participant])
    assert names == {'ALG8', 'CDG1H', 'CDG1C', 'ALG6, CDG1C', 'ALG6',
                     'ALG8, CDG1H, MGC2840'}, names

    # Smoke test serialization
    pybiopax.model_to_owl_file(model, 'test_output.owl')
    model = pybiopax.model_from_owl_file('test_output.owl')
    assert len(model.objects) == 62

    # Test serialization into string
    owl_str = pybiopax.model_to_owl_str(model)
    model = pybiopax.model_from_owl_str(owl_str)
    assert len(model.objects) == 62


def test_get_netpath():
    m = pybiopax.model_from_netpath("22")
    assert isinstance(m, BioPaxModel)
    assert 0 < len(m.objects)


def test_get_reactome():
    m = pybiopax.model_from_reactome("177929")
    assert isinstance(m, BioPaxModel)
    assert re.match("http://www.reactome.org/biopax/\d+/177929#", m.xml_base)
    assert 0 < len(m.objects)


def test_get_humancyc():
    m = pybiopax.model_from_humancyc("PWY66-398")
    assert isinstance(m, BioPaxModel)
    assert m.xml_base is not None
    assert 0 < len(m.objects)


def test_get_biocyc():
    m = pybiopax.model_from_biocyc("P105-PWY")
    assert isinstance(m, BioPaxModel)
    assert m.xml_base is not None
    # This is garbage:
    # 'http://http://BioCyc.org//META/pathway-biopax?type=3%38object=P105-PWY'
    assert 0 < len(m.objects)


def test_get_metacyc():
    m = pybiopax.model_from_metacyc("TCA")
    assert isinstance(m, BioPaxModel)
    assert m.xml_base is not None
    assert 0 < len(m.objects)


def test_get_ecocyc():
    m = pybiopax.model_from_ecocyc("TCA")
    assert isinstance(m, BioPaxModel)
    assert m.xml_base is not None
    assert 0 < len(m.objects)



from collections import Counter
from typing import List, Mapping, Set, Tuple
from .biopax import BioPaxModel, Xref


def get_prefix_id_pairs(model: BioPaxModel) -> List[Tuple[str, str]]:
    """Return a list of database/identifier pairs used in the references of
    a BioPAX Model.

    Parameters
    ----------
    model :
        A BioPAX Model.

    Returns
    -------
    :
        A list of database/identifier pairs used in the model.
    """
    refs = list(model.get_objects_by_type(Xref))
    return [(ref.db, ref.id) for ref in refs]


def get_all_prefixes(model: BioPaxModel) -> Set[str]:
    """Return a set of all prefixes used in the references of a BioPAX Model.

    Parameters
    ----------
    model :
        A BioPAX Model.

    Returns
    -------
    :
        A set of all prefixes used in the model.
    """
    return {db for db, identifier in get_prefix_id_pairs(model)}


def get_prefix_statistics(model: BioPaxModel) -> Mapping[str, int]:
    """Return a dict of prefixes and the number of times they are used in
     references in a BioPAX Model.

    Parameters
    ----------
    model :
        A BioPAX Model.

    Returns
    -------
    :
        A dict of prefixes and the number of times they are used
        in references in the model.
     """
    return dict(Counter([db for db, identifier
                         in get_prefix_id_pairs(model)]).most_common())





import pytest
from pybiopax.biopax import *
from pybiopax import model_from_pc_query
from pybiopax.paths import find_objects, BiopaxClassConstraintError


def test_find_objects():
    protref = EntityReference(uid='1')
    prot = Protein(uid='2', standard_name='x', entity_reference=protref)
    protref._entity_reference_of = [prot]

    # Forward reference
    objects = find_objects(prot, 'entity_reference')
    assert len(objects) == 1, objects
    assert objects[0] == protref

    # Reverse reference
    objects = find_objects(protref, 'entity_reference_of')
    assert len(objects) == 1, objects
    assert objects[0] == prot, objects

    # Multi-step
    objects = find_objects(prot, 'entity_reference/entity_reference_of')
    assert len(objects) == 1, objects
    assert objects[0] == prot, objects


def test_find_objects_constraints():
    xr1 = UnificationXref(uid='2')
    xr2 = PublicationXref(uid='3')
    xr3 = RelationshipXref(uid='4')
    protref = EntityReference(uid='1', xref=[xr1, xr2, xr3])
    xr1._xref_of = [protref]
    xr2._xref_of = [protref]
    xr3._xref_of = [protref]

    objects = find_objects(protref, 'xref')
    assert len(objects) == 3

    objects = find_objects(protref, 'xref:Xref')
    assert len(objects) == 3

    objects = find_objects(protref, 'xref:UnificationXref')
    assert len(objects) == 1
    assert objects[0] == xr1

    objects = find_objects(protref, 'xref:UnificationXref/xref_of')
    assert len(objects) == 1
    assert objects[0] == protref, objects

    objects = find_objects(protref, 'xref:UnificationXref/xref_of:RnaReference')
    assert not objects


def test__with_invalid_class():
    xr1 = UnificationXref(uid='1')

    with pytest.raises(BiopaxClassConstraintError):
        find_objects(xr1, 'xref_of:XXX')


def test_recursive():
    p1 = Protein(uid='1')
    c1 = Complex(uid='2', member_physical_entity=[p1])
    c2 = Complex(uid='3', member_physical_entity=[c1])

    objects = find_objects(c2, 'member_physical_entity')
    assert len(objects) == 1
    assert objects[0] == c1

    objects = find_objects(c2, 'member_physical_entity/member_physical_entity')
    assert len(objects) == 1
    assert objects[0] == p1

    objects = find_objects(c2, 'member_physical_entity*')
    assert len(objects) == 2, objects
    assert set(objects) == {p1, c1}


def test_multi_step():
    model = model_from_pc_query('pathsfromto', ['MAP2K1'], ['MAPK1'])
    bcr = model.objects['BiochemicalReaction_4f689747397d98089c551022a3ae2d88']
    assert len(find_objects(bcr, 'left')) == 1
    assert len(find_objects(bcr, 'left/entity_reference')) == 1
    objs = find_objects(bcr, 'left/entity_reference/entity_reference_of')
    assert len(objs) == 10
